/* mapview - map viewer for PalmOS
   Copyright (C) 2005  Edwin Steiner <edwin.steiner@gmx.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <PalmOS.h>
#include <DataMgr.h>
#include "global.h"
#include "map.h"
#include "memory.h"
#include "debug.h"

typedef struct mapview_rc_map_t mapview_rc_map_t;
typedef struct mapview_rc_layer_t mapview_rc_layer_t;

/* resource structure of map header in prc file */
struct mapview_rc_map_t {
        UInt32  mapsizex;
        UInt32  mapsizey;
        UInt16  nlayers;
        UInt16  density;
};

/* resource structure of layer header in prc file */
struct mapview_rc_layer_t {
        UInt16  scale;
        UInt16  tilesizex;
        UInt16  tilesizey;
        UInt16  tilesx;
        UInt16  tilesy;
        UInt16  tilebaseid;
};

static MemHandle
mapview_map_get_resource(mapview_map_t *map,DmResType type,DmResID id)
{
        UInt16 index;
        
        ASSERT(map);
        
        if (!map->dbref)
                return NULL;
        index = DmFindResource(map->dbref,type,id,NULL);

        return DmGetResourceIndex(map->dbref,index);
}

mapview_map_t *
mapview_map_new(UInt16 cardno,LocalID dbid)
{
        mapview_map_t *map;
        mapview_rc_map_t *maprc;
        MemHandle maprchandle;
        int i;

        map = NEW(mapview_map_t);
        if (!map)
                return NULL;

        map->dbref = DmOpenDatabase(cardno,dbid,dmModeReadOnly);
        if (!map->dbref)
                return NULL;

        maprchandle = mapview_map_get_resource(map,MapviewMapResourceType,MapviewMaprcID);
        if (!maprchandle)
                return NULL;

        maprc = (mapview_rc_map_t *) MemHandleLock(maprchandle);
        if (!maprc)
                return NULL;

        map->mapsizex  = maprc->mapsizex;
        map->mapsizey  = maprc->mapsizey;
        map->nlayers   = maprc->nlayers;
        map->density   = maprc->density;
        map->_mapfile  = NULL;

        ASSERT(map->nlayers > 0);
        
        if (MemHandleUnlock(maprchandle) != errNone)
                return NULL;

        map->layers = NEWARRAY(mapview_layer_t*,map->nlayers);
        if (!map->layers)
                return NULL;

        for (i=0; i<(int)map->nlayers; ++i) {
                map->layers[i] = mapview_layer_new(map,MapviewLayerrcID + i);
                if (!map->layers[i])
                        return NULL;
        }

        return map;
}

mapview_map_t *
mapview_map_new_mapfile(mapview_mapfile_t *mapfile)
{
        mapview_map_t *map;
        mapview_map_t *result;
        mapview_mapfile_header_t *header;
        mapview_mapfile_mapheader_t *mapheader;
        mapview_mapfile_layerheader_t *layerheader;
        int i;

        ASSERT(mapfile);

        result = NULL;
        layerheader = NULL;

        /* allocate the map structure */
        map = NEW(mapview_map_t);
        if (!map)
                return NULL;
        ZERO(mapview_map_t,map);

        /* read the mapfile header */
        header = (mapview_mapfile_header_t *) mapview_mapfile_read_chunk(
                        mapfile,0,MAPVIEW_MAGIC_HEADER);
        if (!header || header->chunk.size < sizeof(mapview_mapfile_header_t)) {
                mapview_message_fmt("Invalid mapfile header");
                goto cleanup_header;
        }

        /* read the map header */
        mapheader = (mapview_mapfile_mapheader_t *) mapview_mapfile_read_chunk(
                        mapfile,header->chunk.size,MAPVIEW_MAGIC_MAPHEADER);
        if (!mapheader || mapheader->chunk.size < sizeof(mapview_mapfile_mapheader_t)) {
                mapview_message_fmt("Invalid map header");
                goto cleanup_mapheader;
        }

        map->dbref    = NULL;
        map->mapsizex = mapheader->mapsizex;
        map->mapsizey = mapheader->mapsizey;
        map->nlayers  = mapheader->nlayers;
        map->density  = mapheader->density;
        map->_mapfile = mapfile;

        /* allocate the layers array */
        map->layers = NEWARRAY(mapview_layer_t*,map->nlayers);
        if (!map->layers)
                goto cleanup_mapheader;
        ZEROARRAY(mapview_layer_t*,map->nlayers,map->layers);

        /* check size of map header */
        if (mapheader->chunk.size < sizeof(mapview_mapfile_mapheader_t) + 4*map->nlayers)
                goto cleanup_mapheader;

        for (i=0; i<(int)map->nlayers; ++i) {
                /* read the layer header */
                layerheader = (mapview_mapfile_layerheader_t *) mapview_mapfile_read_chunk(
                                mapfile,mapheader->layerofs[i],MAPVIEW_MAGIC_LAYERHEADER);
                if (!layerheader || layerheader->chunk.size < sizeof(mapview_mapfile_layerheader_t)) {
                        mapview_message_fmt("Invalid layer header");
                        goto cleanup_layerheader;
                }

                /* create the layer */
                map->layers[i] = mapview_layer_new_mapfile(map,layerheader);
                if (!map->layers[i])
                        goto cleanup_layerheader;

                /* free the layer header */
                mapview_mapfile_free_chunk(&(layerheader->chunk));
                layerheader = NULL;
        }

        /* success */
        result = map;

cleanup_layerheader:
        mapview_mapfile_free_chunk(&(layerheader->chunk));

cleanup_mapheader:
        mapview_mapfile_free_chunk(&(mapheader->chunk));

cleanup_header:
        mapview_mapfile_free_chunk(&(header->chunk));

        if (!result) {
                /* in case of an error free the map structure */
                mapview_map_free(map);
        }
        
        return result;
}

void
mapview_map_free(mapview_map_t *map)
{
        int i;
        
        if (!map)
                return;

        if (map->layers) {
                for (i=0; i<(int)map->nlayers; ++i)
                        mapview_layer_free(map->layers[i]);
                FREEARRAY(mapview_layer_t*,map->nlayers,map->layers);
        }

        if (map->dbref)
                DmCloseDatabase(map->dbref);

        if (map->_mapfile)
                mapview_mapfile_free(map->_mapfile);

        FREE(mapview_map_t,map);
}

mapview_layer_t *
mapview_layer_new(mapview_map_t *map,UInt32 id)
{
        mapview_layer_t *layer;
        mapview_rc_layer_t *layerrc;
        MemHandle rchandle;
        
        ASSERT(map);

        layer = NEW(mapview_layer_t);
        if (!layer)
                return NULL;

        layer->_map = map;

        rchandle = mapview_map_get_resource(map,MapviewLayerResourceType,id);
        if (!rchandle) {
                FREE(mapview_layer_t,layer);
                return NULL;
        }

        layerrc = (mapview_rc_layer_t *) MemHandleLock(rchandle);
        if (!layerrc) {
                FREE(mapview_layer_t,layer);
                return NULL;
        }

        layer->scale       = layerrc->scale;
        layer->tilesizex   = layerrc->tilesizex;
        layer->tilesizey   = layerrc->tilesizey;
        layer->tilesx      = layerrc->tilesx;
        layer->tilesy      = layerrc->tilesy;
        layer->tilebaseid  = layerrc->tilebaseid;
        layer->_tiledirofs = 0;
        layer->_tiledir    = NULL;

        if (MemHandleUnlock(rchandle) != errNone) {
                FREE(mapview_layer_t,layer);
                return NULL;
        }

        layer->_bitmaphandle = NULL;

        return layer;
}

mapview_layer_t *
mapview_layer_new_mapfile(mapview_map_t *map,mapview_mapfile_layerheader_t *layerheader)
{
        mapview_layer_t *layer;
        
        ASSERT(map);
        ASSERT(layerheader);

        layer = NEW(mapview_layer_t);
        if (!layer)
                return NULL;

        layer->scale         = layerheader->scale;
        layer->tilesizex     = layerheader->tilesizex;
        layer->tilesizey     = layerheader->tilesizey;
        layer->tilesx        = layerheader->tilesx;
        layer->tilesy        = layerheader->tilesy;
        layer->tilebaseid    = 0;
        layer->_tiledirofs   = layerheader->tiledirofs;
        layer->_tiledir      = NULL;
        layer->_map          = map;
        layer->_bitmaphandle = NULL;

        return layer;
}

void
mapview_layer_free(mapview_layer_t *layer)
{
        if (!layer)
                return;

        if (layer->_bitmaphandle != NULL)
                mapview_layer_release_tile(layer);

        if (layer->_tiledir)
                mapview_mapfile_free_chunk(&(layer->_tiledir->chunk));

        FREE(mapview_layer_t,layer);
}

static bool
mapview_layer_load_tiledir(mapview_layer_t *layer)
{
        mapview_mapfile_tiledir_t *tiledir;
        
        ASSERT(layer);
        ASSERT(layer->_tiledirofs != 0);
        ASSERT(layer->_map);
        ASSERT(layer->_map->_mapfile);
        ASSERT(layer->_tiledir == NULL);

        tiledir = (mapview_mapfile_tiledir_t *)mapview_mapfile_read_chunk(
                        layer->_map->_mapfile,
                        layer->_tiledirofs,MAPVIEW_MAGIC_TILEDIR);
        if (!tiledir)
                return false;

        layer->_tiledir = tiledir;

        return true;
}

BitmapPtr 
mapview_layer_get_tile(mapview_layer_t *layer,Int32 tx,Int32 ty)
{
        ASSERT(layer);
        ASSERT(layer->_bitmaphandle == NULL);

        if (tx < 0 || ty < 0 || tx >= layer->tilesx || ty >= layer->tilesy)
                return NULL;
        
        if (layer->_tiledirofs) {
                UInt32 ofs;
                mapview_mapfile_tile_t *tile;

                /* get a tile from a VFS mapfile */
                /* first we need the tile directory of this layer */
                if (!layer->_tiledir)
                        if (!mapview_layer_load_tiledir(layer))
                                        return NULL;

                ofs = layer->_tiledir->tileofs[ty*layer->tilesx + tx];
                if (!ofs)
                        return NULL;

                /* read the tile from the file */
                tile = (mapview_mapfile_tile_t *) mapview_mapfile_read_chunk(
                                layer->_map->_mapfile,
                                ofs,
                                MAPVIEW_MAGIC_TILE);
                if (!tile)
                        return NULL;

                layer->_bitmaphandle = (MemHandle) tile;

                return (BitmapPtr) &(tile->data);
        }
        else {
                UInt32 id;

                /* get a tile from a resource database */
                id = (UInt32)layer->tilebaseid + (ty*layer->tilesx + tx);

                layer->_bitmaphandle = mapview_map_get_resource(layer->_map,bitmapRsc,id);
                if (!layer->_bitmaphandle)
                        return NULL;

                return (BitmapPtr) MemHandleLock(layer->_bitmaphandle);
        }
}

void
mapview_layer_release_tile(mapview_layer_t *layer)
{
        ASSERT(layer);
        ASSERT(layer->_bitmaphandle != NULL);

        if (layer->_tiledirofs) {
                /* the tile is from a mapfile */
                mapview_mapfile_free_chunk((mapview_mapfile_chunk_t *)layer->_bitmaphandle);
        }
        else {
                /* the tile is from a resource database */
                MemHandleUnlock(layer->_bitmaphandle);
        }
        layer->_bitmaphandle = NULL;
}


/* vim: et sts=8 sw=8 ai
 */
