/* mapview - map viewer for PalmOS
   Copyright (C) 2005  Edwin Steiner <edwin.steiner@gmx.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <PalmOS.h>
#include <VFSMgr.h>
#include <string.h>
#include "mapfile.h"
#include "memory.h"
#include "debug.h"

mapview_mapfile_t *
mapview_mapfile_open(UInt16 vol,const char *path)
{
        mapview_mapfile_t *mapfile;
        FileRef fileref;
        Err err;
    
        ASSERT(path);

        mapfile = NEW(mapview_mapfile_t);
        if (!mapfile)
                return NULL;

        err = VFSFileOpen(vol,path,vfsModeRead,&fileref);
        if (err != errNone) {
                FREE(mapview_mapfile_t,mapfile);
                mapview_message_fmt("VFSFileOpen failed: %ld",(long)err);
                return NULL;
        }

        mapfile->fileref = fileref;

        return mapfile;
}

void
mapview_mapfile_free(mapview_mapfile_t *mapfile)
{
        Err err;
        
        if (!mapfile)
                return;

        err = VFSFileClose(mapfile->fileref);
        if (err != errNone) {
                /* supress error message as it may be irritating for users */
                /* and the error really can be ignored.                    */
                /*mapview_message_fmt("VFSFileClose failed: %ld",(long)err);*/
        }

        FREE(mapview_mapfile_t,mapfile);
}

mapview_mapfile_chunk_t *
mapview_mapfile_read_chunk(mapview_mapfile_t *mapfile,UInt32 ofs,UInt32 magic)
{
        mapview_mapfile_chunk_t chunk;
        Err err;
        UInt32 bytes;
        UInt32 toread;
        UInt8 *data;

        ASSERT(mapfile);

        /* seek the given offset */
        err = VFSFileSeek(mapfile->fileref,vfsOriginBeginning,(Int32)ofs);
        if (err != errNone) {
                mapview_message_fmt("VFSFileSeek failed: %ld",(long)err);
                return NULL;
        }

        /* read the chunk header */
        err = VFSFileRead(mapfile->fileref,sizeof(chunk),&chunk,&bytes);
        if (err != errNone) {
                mapview_message_fmt("VFSFileRead (1) failed: %ld",(long)err);
                return NULL;
        }
        if (bytes != sizeof(chunk)) {
                mapview_message_fmt("VFSFileRead (1) incomplete");
                return NULL;
        }
        
        /* verify magic number and size of the chunk header */
        if (magic != MAPVIEW_MAGIC_ANY && chunk.magic != magic) {
                mapview_message_fmt("Wrong magic: %08lx. Expected: %08lx",
                                (unsigned long)chunk.magic,(unsigned long)magic);
                return NULL;
        }
        if (chunk.size < sizeof(chunk)) {
                mapview_message_fmt("Chunk too small");
                return NULL;
        }

        /* allocate space for the chunk */
        toread = chunk.size;
        data = NEWARRAY(UInt8,toread);
        if (!data)
                return NULL;

        /* copy the chunk header we already read */
        memcpy(data,&chunk,sizeof(chunk));
        toread -= sizeof(chunk);

        /* read the rest of the chunk */
        err = VFSFileRead(mapfile->fileref,toread,data + sizeof(chunk),&bytes);
        if (err != errNone) {
                mapview_message_fmt("VFSFileRead (2) failed: %ld",(long)err);
                goto error_free;
        }
        if (bytes != toread) {
                mapview_message_fmt("VFSFileRead (2) incomplete");
                goto error_free;
        }

        /* success */
        return (mapview_mapfile_chunk_t *)data;

error_free:
        FREEARRAY(UInt8,chunk.size,data);
        return NULL;
}

void 
mapview_mapfile_free_chunk(mapview_mapfile_chunk_t *chunk)
{
        if (!chunk)
                return;

        FREEARRAY(UInt8,chunk->size,chunk);
}

/* vim: et sts=8 sw=8 ai
 */
